<?php
/*
* Plugin Name: CodeFusion Image Optimizer
* Description: Optimizes images and scans them for post types and media library with API validation and progress tracking.
* Version: 1.1
* Author: codefusiononline
* Author URI:  https://codefusiononline.com/
* License: GPLv3 or later
* Text Domain: CodeFImageOptimizer
*/

if ( ! defined( 'ABSPATH' ) ) {
    exit; // Exit if accessed directly
}

require plugin_dir_path(__FILE__) . 'update/plugin-update-checker.php';
use YahnisElsts\PluginUpdateChecker\v5\PucFactory;

$myUpdateChecker = PucFactory::buildUpdateChecker(
    'https://codefusiononline.com/updates/codefusiononline.json', // JSON file URL
    __FILE__, // Plugin's main file path
    'CodeFImageOptimizer' // Plugin slug
);

class CodeF_Image_Optimizer {
    private $app_url = 'https://codefusiononline.com';
    private $api_url = 'https://codefusiononline.com/optimize-image';
    private $checkapikey = 'https://codefusiononline.com/api/checkapikey';
    private $upload_dir;
    private $plugin_dir;

    public function __construct() {
        $this->upload_dir = wp_upload_dir()['basedir'] . '/cfonline/';
        $this->plugin_dir = plugin_dir_path( __FILE__ );

        register_activation_hook( __FILE__, array( $this, 'activate' ) );
        register_deactivation_hook( __FILE__, array( $this, 'deactivate' ) );

        add_action( 'admin_menu', array( $this, 'add_admin_menu' ) );
        add_action( 'admin_init', array( $this, 'register_settings' ) );
        add_action( 'admin_enqueue_scripts', array( $this, 'enqueue_scripts' ) );
        add_action( 'codef_optimize_images_cron', array( $this, 'optimize_images' ) );
		// Register the AJAX action
		add_action('wp_ajax_start_image_optimization', array( $this, 'start_image_optimization_callback'));
		add_action('wp_ajax_start_scan_image', array( $this, 'start_scan_image_callback'));
		add_action('wp_ajax_load_optimized_images', array( $this, 'load_optimized_images'));
		
		add_filter('wp_get_attachment_image_src', array( $this, 'modify_image_path_and_extension'), 10, 4);
		add_filter('wp_calculate_image_srcset', array($this, 'modify_srcset_image_path'), 10, 5);
		add_filter('the_content', array($this, 'modify_image_paths_in_content'), 10, 1);	
    }	
	
	public function modify_srcset_image_path($sources, $size_array, $image_src, $image_meta, $attachment_id) {
		if (!empty($sources)) {
			foreach ($sources as $key => $source) {
				$new_url = str_replace('/uploads', '/uploads/cfonline', $source['url']);
				$new_url = preg_replace('/\.(jpg|jpeg|png)$/i', '.webp', $new_url);

				// Check if the file exists
				$file_path = ABSPATH . str_replace(get_site_url(), '', $new_url);
				if (file_exists($file_path)) {
					$sources[$key]['url'] = $new_url;
				}
			}
		}
		return $sources;
	}

	public function modify_image_paths_in_content($content) {
		// Replace "/uploads" with "/uploads/cfonline" and update to .webp
		$content = preg_replace_callback(
			'/<img[^>]+src=["\']([^"\']+)["\'][^>]*>/i',
			function ($matches) {
				$img_tag = $matches[0];
				$img_url = $matches[1];

				// Replace "/uploads" with "/uploads/cfonline"
				$new_url = str_replace('/uploads', '/uploads/cfonline', $img_url);

				// Replace ".jpg", ".jpeg", or ".png" with ".webp"
				$new_url = preg_replace('/\.(jpg|jpeg|png)$/i', '.webp', $new_url);

				// Check if the file exists
				$file_path = ABSPATH . str_replace(get_site_url(), '', $new_url);
				if (file_exists($file_path)) {
					// Replace old src URL with new URL in the <img> tag
					return str_replace($img_url, $new_url, $img_tag);
				}

				return $img_tag; // Return the original tag if the file doesn't exist
			},
			$content
		);

		return $content;
	}

	public function modify_image_path_and_extension($image, $attachment_id, $size, $icon) {
		if (!empty($image) && is_array($image)) {
			$new_url = str_replace('/uploads', '/uploads/cfonline', $image[0]);
			$new_url = preg_replace('/\.(jpg|jpeg|png)$/i', '.webp', $new_url);

			// Check if the file exists
			$file_path = ABSPATH . str_replace(get_site_url(), '', $new_url);
			if (file_exists($file_path)) {
				$image[0] = $new_url;
			}
		}
		return $image;
	}

	
	public function load_optimized_images() {
		global $wpdb;
		$page = isset($_GET['page']) ? (int)$_GET['page'] : 1;
		$items_per_page = isset($_GET['items_per_page']) ? (int)$_GET['items_per_page'] : 10;

		ob_start();
			$this->display_optimized_images($page, $items_per_page);
		$html = ob_get_clean();

		wp_send_json_success(['html' => $html]);
	}
	
	public function start_scan_image_callback() {
		// Run your image optimization logic and return the response
		$optimized = $this->scan_images(); // Replace with your optimization function

		if ($optimized) {
			wp_send_json_success(['message' => 'Images optimized successfully.']);
		} else {
			wp_send_json_error(['message' => 'Image optimization failed.']);
		}
	}
	
	public function start_image_optimization_callback() {
		// Run your image optimization logic and return the response
		$optimized = $this->optimize_images(); // Replace with your optimization function

		if ($optimized) {
			wp_send_json_success(['message' => 'Images optimized successfully.']);
		} else {
			wp_send_json_error(['message' => 'Image optimization failed.']);
		}
	}

    public function activate() {
        $this->create_database_table();
        if ( ! wp_next_scheduled( 'codef_optimize_images_cron' ) ) {
            wp_schedule_event( time(), 'hourly', 'codef_optimize_images_cron' );
        }
        if ( ! file_exists( $this->upload_dir ) ) {
            wp_mkdir_p( $this->upload_dir );
        }
    }

    public function deactivate() {
        wp_clear_scheduled_hook( 'codef_optimize_images_cron' );
    }

    private function create_database_table() {
        global $wpdb;
        $table_name = $wpdb->prefix . 'codef_optimized_images';
        $charset_collate = $wpdb->get_charset_collate();

        $sql = "CREATE TABLE IF NOT EXISTS $table_name (
            id INT NOT NULL AUTO_INCREMENT,
            original_url TEXT NOT NULL,
            optimized_url TEXT NOT NULL,
            status VARCHAR(20) NOT NULL,
            PRIMARY KEY (id)
        ) $charset_collate;";

        require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );
        dbDelta( $sql );
    }

    public function add_admin_menu() {
        add_menu_page( 'CFOnline Image Optimizer', 'CFOnline Images', 'manage_options', 'codef_image_optimizer', array( $this, 'render_admin_page' ), 'dashicons-format-image' );
    }

    public function register_settings() {
        register_setting( 'codef_image_optimizer', 'codef_api_key' );
		register_setting( 'codef_image_optimizer', 'codef_image_format' );
		register_setting( 'codef_image_optimizer', 'codef_image_quality' );
		register_setting( 'codef_options_group', 'codef_max_height' );
		register_setting( 'codef_options_group', 'codef_max_width' );
	
        add_settings_section( 'codef_section', 'Settings', null, 'codef_image_optimizer' );
        add_settings_field( 'codef_api_key', 'API Key', array( $this, 'api_key_field_callback' ), 'codef_image_optimizer', 'codef_section' );
		add_settings_field( 'codef_image_format', 'Image Format', array( $this, 'image_format_field_callback' ), 'codef_image_optimizer', 'codef_section' );
		add_settings_field( 'codef_image_quality', 'Image Quality', array( $this, 'image_quality_field_callback' ), 'codef_image_optimizer', 'codef_section' );
		add_settings_field( 'codef_max_width', 'Max Width', array( $this, 'max_width_field_callback' ), 'codef_image_optimizer', 'codef_section' );
		add_settings_field( 'codef_max_height', 'Max Height', array( $this, 'max_height_field_callback' ), 'codef_image_optimizer', 'codef_section' );
    }

    public function api_key_field_callback() {
        $api_key = get_option( 'codef_api_key' );
		
		
        $status = get_option( 'codef_api_key_status' );
        echo '<input type="text" name="codef_api_key" value="' . esc_attr( $api_key ) . '" class="regular-text">';
		echo '<span id="api-key-status">';
		
				
		// Display validation result if available
		if (isset($api_key)) {
			$this->validate_api_key($api_key);
			//$this->update_htaccess();
		}
		
		echo '</span>';
    }
	
	public function validate_api_key($api_key) {
		update_option('codef_api_key_status', 0);
		// Set up the remote request
		/*$response = wp_remote_post($this->checkapikey, [
			'headers' => [
				'Content-Type' => 'application/json',
				'x-api-key' => $api_key,
			],
			'body' => json_encode(['api_key' => $api_key]),
		]);

		// Check for errors
		if (is_wp_error($response)) {
			echo '<span style="color: red;">Error validating API key.</span>';
			return;
		}

		// Decode and handle the response
		$body = wp_remote_retrieve_body($response);
		$data = json_decode($body);

		if (isset($data->error)) {
			echo '<span style="color: red;">Invalid API key.</span>';
		} else {
			update_option('codef_api_key_status', 1);
			echo '<span style="color: green;"><span class="dashicons dashicons-yes-alt"></span></span>';
		}*/
	}

	
	public function max_width_field_callback() {
        $val = get_option( 'codef_max_width' );
        echo '<input type="text" name="codef_max_width" value="' . esc_attr( $val ) . '" class="regular-text">';
    }
	
	public function max_height_field_callback() {
        $val = get_option( 'codef_max_height' );
        echo '<input type="text" name="codef_max_height" value="' . esc_attr( $val ) . '" class="regular-text">';
    }
	
	public function image_format_field_callback() {
		$format = get_option( 'codef_image_format', 'webp' );
		echo '<select name="codef_image_format">
            <option value="jpeg"' . selected( $format, 'jpeg', false ) . '>JPEG (.jpeg, .jpg)</option>
            <option value="png"' . selected( $format, 'png', false ) . '>PNG (.png)</option>
            <option value="webp"' . selected( $format, 'webp', false ) . '>WebP (.webp)</option>
            <option value="gif"' . selected( $format, 'gif', false ) . '>GIF (input only; GIFs are flattened to a static image)</option>
            <option value="tiff"' . selected( $format, 'tiff', false ) . '>TIFF (.tiff, .tif)</option>
            <option value="avif"' . selected( $format, 'avif', false ) . '>AVIF (.avif)</option>
            <option value="heif"' . selected( $format, 'heif', false ) . '>HEIF/HEIC (input only; common for photos on iOS devices)</option>            
            <option value="raw"' . selected( $format, 'raw', false ) . '>RAW (input only; various raw image formats from cameras)</option>
            <option value="bmp"' . selected( $format, 'bmp', false ) . '>BMP (.bmp)</option>
          </select>';
	}

	public function image_quality_field_callback() {
		$quality = get_option( 'codef_image_quality', 80 );
		echo '<input type="number" name="codef_image_quality" value="' . esc_attr( $quality ) . '" min="10" max="100" class="small-text"> %';
	}
	
	private function update_htaccess() {
		$htaccess_path = ABSPATH . '.htaccess';

		// The custom rules for image optimization
		$htaccess_rules = "
						# BEGIN cfonline Image Optimizer
						<IfModule mod_rewrite.c>
						RewriteEngine On
						RewriteCond %{REQUEST_FILENAME} -f
						RewriteCond %{REQUEST_FILENAME} \\.(jpg|jpeg|png|gif|bmp|webp|tiff)$ [NC]
						RewriteRule ^wp-content/uploads/(.*)$ " . wp_upload_dir()['baseurl'] . "/cfonline/$1 [L]
						</IfModule>
						# END cfonline Image Optimizer
						";

		// Check if .htaccess exists and is writable
		if (file_exists($htaccess_path) && is_writable($htaccess_path)) {
			$current_htaccess = file_get_contents($htaccess_path);

			// If the rules are already present, replace them; if not, add them
			if (strpos($current_htaccess, '# BEGIN cfonline Image Optimizer') === false) {
				// Rules do not exist, so add them
				file_put_contents($htaccess_path, $current_htaccess . PHP_EOL . $htaccess_rules);
			} else {
				// Rules exist, replace with updated rules
				$updated_htaccess = preg_replace(
					'/# BEGIN cfonline Image Optimizer.*?# END cfonline Image Optimizer/s',
					$htaccess_rules,
					$current_htaccess
				);
				file_put_contents($htaccess_path, $updated_htaccess);
			}
		} else {
			// Log error if .htaccess is not writable
			error_log('Error: .htaccess file is not writable. Please check file permissions.');
		}
	}



    public function enqueue_scripts( $hook ) {
        if ( $hook !== 'toplevel_page_codef_image_optimizer' ) return;
        wp_enqueue_style( 'bootstrap-css',  plugins_url( 'assert/css/bootstrap.min.css', __FILE__ ));
        wp_enqueue_script( 'codef-image-optimizer', plugins_url( 'assert/js/codef-image-optimizer.js', __FILE__ ), array( 'jquery' ), null, true );
        wp_localize_script( 'codef-image-optimizer', 'codef_optimizer', array(
            'ajax_url' => admin_url( 'admin-ajax.php' ),
        ) );
    }

    public function render_admin_page() {
        ?>
        <div class="wrap">
            <h1>CFOnline Image Optimizer</h1>
            <form method="post" action="options.php">
                <?php
                settings_fields( 'codef_image_optimizer' );
                do_settings_sections( 'codef_image_optimizer' );
                submit_button();
                ?>
            </form>

            <h2>Image Optimization</h2>
            <button id="codef_scan_images" class="button button-primary">Scan for Images</button>
            <button id="codef_optimize_images" class="button button-secondary">Run Optimization</button>
            <div id="codef_results" class="mt-3"></div>

            <h2>Optimized Images</h2>
            <table id ="optimized-images-table-body" class="table table-bordered mt-3">
                <thead>
                    <tr>
                        <th>#</th>
                        <th>Image URL</th>
                        <th>Optimized URL</th>
                        <th>Status</th>
                    </tr>
                </thead>
                <tbody>
                    <?php $this->display_optimized_images(); ?>
                </tbody>
            </table>
			<!-- Pagination Controls -->
			<div class="pagination-controls mt-3">
				<button id="previous-page" class="btn btn-secondary">Previous</button>
				<span id="current-page">Page 1</span>
				<button id="next-page" class="btn btn-secondary">Next</button>
			</div>
        </div>
        <?php
    }

    // Display paginated optimized images
	public function display_optimized_images($page = 1, $items_per_page = 10) {
		global $wpdb;
		$table_name = $wpdb->prefix . 'codef_optimized_images';
		$offset = ($page - 1) * $items_per_page;

		// Fetch paginated results
		$images = $wpdb->get_results($wpdb->prepare(
			"SELECT * FROM $table_name LIMIT %d OFFSET %d", 
			$items_per_page, 
			$offset
		));

		foreach ($images as $index => $image) {
			echo '<tr>
					<td>' . esc_html($index + 1 + $offset) . '</td>
					<td>' . esc_url($image->original_url) . '</td>
					<td>' . esc_url($image->optimized_url) . '</td>
					<td>' . esc_html($image->status) . '</td>
				  </tr>';
		}
	}

	// Scan images and insert non-existing entries
	public function scan_images() {
		global $wpdb;
		try {
			$table_name = $wpdb->prefix . 'codef_optimized_images';
			$images = array_merge($this->get_images(), $this->get_theme_images());
			$api_key = get_option('codef_api_key');
			$site_url = parse_url(get_site_url());

			if (!$api_key) return;

			foreach ($images as $image) {
				$image_url = is_object($image) && isset($image->guid) ? $image->guid : (is_string($image) ? $image : null);
				if (!$image_url) continue;

				// Check if the URL belongs to the site
				$parsed_image_url = parse_url($image_url);
				if ($parsed_image_url['host'] !== $site_url['host'] || strpos($parsed_image_url['path'], $site_url['path']) !== 0) {
					continue;
				}

				$image_url_path = parse_url($image_url, PHP_URL_PATH); // Gets "/wp-content/uploads/2023/11/woocommerce-placeholder.webp"

				// Remove the initial part of the path that matches the original upload directory structure
				$image_path_relative = str_replace('/wp-content/uploads', '', $image_url_path);
				// Construct the optimized image path by combining the new upload directory and the relative path
				$optimized_image_path = $this->upload_dir . ltrim($image_path_relative, '/');

				// Change the file extension to the desired format
				$format = get_option('codef_image_format', 'webp');
				$optimized_image_path = preg_replace('/\.[^.\s]{3,4}$/', '.' . $format, $optimized_image_path);

				if (!file_exists($optimized_image_path)) {
					$exists = $wpdb->get_var($wpdb->prepare(
						"SELECT COUNT(*) FROM $table_name WHERE original_url = %s", 
						$image_url
					));

					if (!$exists) {
						$wpdb->insert($table_name, [
							'original_url' => $image_url,
							'optimized_url' => '',
							'status' => 'Not Optimized'
						]);
					}
				}
			}

			wp_send_json_success(['message' => 'Scanning complete!']);
		} catch (Exception $e) {
			wp_send_json_error(['message' => 'Scan failed: ' . $e->getMessage()]);
		}
		wp_die();
	}

	// Optimize images and update database
	public function optimize_images() {
		global $wpdb;
		try {
			$table_name = $wpdb->prefix . 'codef_optimized_images';
			$images = array_merge($this->get_images(), $this->get_theme_images());
			$api_key = get_option('codef_api_key');
			$site_url = parse_url(get_site_url());
			
			

			if (!$api_key) return;

			foreach ($images as $image) {
				$image_url = is_object($image) && isset($image->guid) ? $image->guid : (is_string($image) ? $image : null); 
				if (!$image_url) continue;

				$parsed_image_url = parse_url($image_url);
				if ($parsed_image_url['host'] !== $site_url['host'] || strpos($parsed_image_url['path'], $site_url['path']) !== 0) {
					continue;
				}

				$image_url_path = parse_url($image_url, PHP_URL_PATH); // Gets "/wp-content/uploads/2023/11/woocommerce-placeholder.webp"

				// Remove the initial part of the path that matches the original upload directory structure
				$image_path_relative = str_replace('/wp-content/uploads', '', $image_url_path);
				// Construct the optimized image path by combining the new upload directory and the relative path
				$optimized_image_path = $this->upload_dir . ltrim($image_path_relative, '/');

				// Change the file extension to the desired format
				$format = get_option('codef_image_format', 'webp');
				$optimized_image_path = preg_replace('/\.[^.\s]{3,4}$/', '.' . $format, $optimized_image_path);
				
				if (!file_exists($optimized_image_path)) {
					$optimized_url = $this->optimize_image($image_url, $optimized_image_path, $api_key);

					if ($optimized_url) {
						$exists = $wpdb->get_var($wpdb->prepare(
							"SELECT COUNT(*) FROM $table_name WHERE original_url = %s", 
							$image_url
						));

						if ($exists) {
							$wpdb->update($table_name, [
								'optimized_url' => $optimized_url,
								'status' => 'Optimized'
							], ['original_url' => $image_url]);
						} else {
							$wpdb->insert($table_name, [
								'original_url' => $image_url,
								'optimized_url' => $optimized_url,
								'status' => 'Optimized'
							]);
						}
					}
				}
			}

			wp_send_json_success(['message' => 'Optimization complete!']);
		} catch (Exception $e) {
			wp_send_json_error(['message' => 'Optimization failed: ' . $e->getMessage()]);
		}
		wp_die();
	}



    private function get_images() {
        global $wpdb;
        return $wpdb->get_results( "SELECT ID, guid FROM {$wpdb->posts} WHERE post_type = 'attachment' AND post_mime_type LIKE 'image/%'" );
    }
	
	private function get_theme_images() {
		$images = array();
		$theme_dirs = array( get_stylesheet_directory(), get_template_directory() );
		
		foreach ( $theme_dirs as $dir ) {
			$files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator($dir));
			foreach ($files as $file) {
				if (in_array(strtolower($file->getExtension()), array('jpg', 'jpeg', 'png', 'gif', 'webp'))) {
					$images[] = $file->getPathname();
				}
			}
		}
		return $images;
	}

    public function optimize_image( $image_url, $save_path, $api_key ) {
		$format = get_option( 'codef_image_format', 'webp' ); // Default to webp
		$quality = get_option( 'codef_image_quality', 80 ); // Default quality
		$max_height = get_option( 'codef_max_height'); // Default max height
		$max_width = get_option( 'codef_max_width'); // Default max width	
		
		$args = array(
            'imageUrl' => $image_url,          
            'format'   => $format,
            'quality'  => $quality,
            'apikey'  => $api_key,
        );
				
		if(isset($max_width)){
			$args['maxwidth'] = $max_width;
		}
		
		if(isset($max_height)){
			$args['maxheight'] = $max_height;
		}
		
		// Build the complete URL for the GET request
		$url = add_query_arg($args, $this->api_url);
		// Send the GET request
		$response = wp_remote_get($url);

		if (is_wp_error($response)) {
			error_log('Error optimizing image: ' . $response->get_error_message());
			return false;
		}

		$optimized_image_data = wp_remote_retrieve_body( $response );
		
		// Check if the response is valid JSON
		$decoded_response = json_decode($optimized_image_data, true);
		
		if (json_last_error() === JSON_ERROR_NONE) {
			// Check for errors in the decoded JSON response
			if (isset($decoded_response['error'])) {
				error_log('Image optimization error: ' . $decoded_response['error']);
				return false; // Do not save if there is an error in the response
			}
		} else {
			if ( ! empty( $optimized_image_data ) ) {	
				
				// Ensure the directory exists
				$directory = dirname($save_path);
				if (!is_dir($directory)) {
					mkdir($directory, 0755, true); // Create directory if it doesn't exist
				}

				file_put_contents( $save_path, $optimized_image_data );
				$image_url = preg_replace('/\.[^.\s]{3,4}$/', '.' . $format, $image_url);
				$image_url_path = parse_url($image_url, PHP_URL_PATH); // Gets "/wp-content/uploads/2023/11/woocommerce-placeholder.webp"

				// Remove the initial part of the path that matches the original upload directory structure
				$optimized_image_path = str_replace($this->upload_dir, '/', $image_url_path);
				// Construct the optimized image path by combining the new upload directory and the relative path
				//$optimized_image_path = $this->upload_dir . ltrim($image_path_relative, '/');

				// Change the file extension to the desired format
				$format = get_option('codef_image_format', 'webp');
				$optimized_image_path = preg_replace('/\.[^.\s]{3,4}$/', '.' . $format, $optimized_image_path);
				return $optimized_image_path;
			}
		}

		return false;
	}
}

new CodeF_Image_Optimizer();
?>
